# NEON AI (TM) SOFTWARE, Software Development Kit & Application Development System
# All trademark and other rights reserved by their respective owners
# Copyright 2008-2021 Neongecko.com Inc.
# BSD-3
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from this
#    software without specific prior written permission.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS;  OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE,  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import os.path
import yaml

from typing import Optional
from os import getenv
from os.path import dirname, join, expanduser, isdir, isfile
from ovos_utils.log import LOG


def cli_make_rmq_config_map(input_path: str, output_path: str) -> str:
    """
    Generate a ConfigMap object for RabbitMQ from general config files
    :param input_path: path to directory containing RabbitMQ config files
    :param output_path: file or dir to write Kubernetes spec file to
    :returns: path to output config file
    """
    # TODO: Deprecate Method
    LOG.warning("This function is deprecated")
    file_path = expanduser(input_path)
    if not isdir(file_path):
        raise FileNotFoundError(f"Could not find requested directory: "
                                f"{input_path}")
    output_path = expanduser(output_path)
    if isdir(output_path):
        output_file = join(output_path, f"k8s_config_rabbitmq.yml")
    elif isfile(output_path):
        output_file = output_path
    else:
        raise ValueError(f"Invalid output_path: {output_path}")

    with open(join(file_path, "rabbitmq.conf"), 'r') as f:
        rabbitmq_file_contents = f.read()
    with open(join(file_path, "rabbit_mq_config.json")) as f:
        rmq_config = f.read()
    generate_config_map("rabbitmq", {"rabbitmq.conf": rabbitmq_file_contents,
                                     "rabbit_mq_config.json": rmq_config}, output_file)
    return output_file


def cli_make_api_secret(input_path: str, output_path: str) -> str:
    """
    Generate a Secret object for ngi-auth from general config files
    :param input_path: path to directory containing ngi_auth_vars.yml
    :param output_path: file or dir to write Kubernetes spec file to
    :returns: path to output config file
    """
    # TODO: Deprecate Method
    LOG.warning("This function is deprecated")
    file_path = expanduser(input_path)
    if not isdir(file_path):
        raise FileNotFoundError(f"Could not find requested directory: {input_path}")
    output_path = expanduser(output_path)
    if isfile(output_path):
        output_path = dirname(output_path)

    with open(join(file_path, "ngi_auth_vars.yml")) as f:
        ngi_auth = f.read()
    generate_secret("ngi-auth", {"ngi_auth_vars.yml": ngi_auth},
                    join(output_path, "k8s_secret_ngi-auth.yml"))
    return output_path


def cli_make_github_secret(username: str, token: str, config_dir: str) -> str:
    """
    Generate a Secret object for a GitHub image pull secret
    :param username: Github username
    :param token: Github PAT with package_read permissions
    :param config_dir: Directory to write output spec file to
    :returns: path to output config file
    """
    output_file = _create_github_secret(username, token,
                                        join(config_dir,
                                             "k8s_secret_github.yml"))
    return output_file


def write_kubernetes_spec(k8s_config: list, output_path: Optional[str] = None,
                          namespaces: dict = None):
    """
    Generates and writes a kubernetes.yml spec file according to the passed services
    :param k8s_config: list of k8s objects specified, usually read from service_mappings.yml
    :param output_path: path to write spec files to
    :param namespaces: dict of placeholders to namespaces
    """
    # TODO: Deprecate Method
    LOG.warning("This function is deprecated")
    namespaces = namespaces or dict()
    output_dir = expanduser(output_path) if output_path else \
        expanduser(getenv("NEON_CONFIG_PATH", "~/.config/neon"))

    diana_spec_file = join(output_dir, "services", "k8s_diana_backend.yml")
    os.makedirs(dirname(diana_spec_file), exist_ok=True)

    # Write Diana services spec file
    with open(join(dirname(dirname(__file__)), "templates",
                   "kubernetes.yml")) as f:
        diana_spec_contents = yaml.safe_load(f)
    diana_spec_contents["items"].extend(k8s_config)

    with open(diana_spec_file, "w+") as f:
        yaml.dump(diana_spec_contents, f)
        f.seek(0)
        string_contents = f.read()
        for placeholder, replacement in namespaces.items():
            string_contents = string_contents.replace(
                '${' + placeholder + '}', replacement)
        f.seek(0)
        f.truncate(0)
        f.write(string_contents)

    # Write Ingress spec file
    os.makedirs(join(output_path, "ingress"), exist_ok=True)

    tcp_config = {'5672': f"{namespaces.get('MQ_NAMESPACE') or 'default'}"
                          f"/neon-rabbitmq:5672"}
    tcp_config = _update_tcp_config(tcp_config,
                                    join(output_path, "ingress",
                                         "k8s_config_tcp_services.yml"))
    LOG.info(f"Wrote {tcp_config}")
    ingress_config = _patch_ingress_nginx_controller_service(
        "neon-rabbitmq", 5672,
        output_path=join(output_path, "ingress",
                         "k8s_patch_nginx_service.yml"))
    LOG.info(f"Wrote {ingress_config}")


def generate_config_map(name: str, config_data: dict, output_path: str):
    """
    Generate a Kubernetes ConfigMap yml definition
    :param name: ConfigMap name
    :param config_data: Dict data to store
    :param output_path: output file to write
    """
    # TODO: Deprecate Method
    LOG.warning("This function is deprecated")
    output_path = output_path or join(getenv("NEON_CONFIG_PATH", "~/.config/neon"), f"k8s_config_{name}.yml")
    output_path = expanduser(output_path)
    if not isdir(dirname(output_path)):
        os.makedirs(dirname(output_path), exist_ok=True)
    config_template = join(dirname(dirname(__file__)),
                           "templates", "k8s_config_map.yml")
    with open(config_template) as f:
        config_map = yaml.safe_load(f)

    config_map["metadata"]["name"] = name
    config_map["data"] = config_data

    with open(output_path, 'w+') as f:
        yaml.dump(config_map, f)


def generate_secret(name: str, secret_data: dict,
                    output_path: Optional[str] = None):
    """
    Generate a Kubernetes Secret yml definition
    :param name: ConfigMap name
    :param secret_data: Dict data to store
    :param output_path: output file to write
    """
    # TODO: Deprecate Method
    LOG.warning("This function is deprecated")
    output_path = output_path or join(getenv("NEON_CONFIG_PATH",
                                             "~/.config/neon"),
                                      f"k8s_secret_{name}.yml")
    output_path = expanduser(output_path)
    if not isdir(dirname(output_path)):
        os.makedirs(dirname(output_path), exist_ok=True)
    config_template = join(dirname(dirname(__file__)),
                           "templates", "k8s_secret.yml")
    with open(config_template) as f:
        config_map = yaml.safe_load(f)

    config_map["metadata"]["name"] = name
    config_map["stringData"] = secret_data

    with open(output_path, 'w+') as f:
        yaml.dump(config_map, f)


def _create_github_secret(username: str, token: str,
                          output_path: Optional[str] = None) -> str:
    """
    Generate a Kubernetes Secret to authenticate to GitHub for image pulls
    :param username: GitHub username
    :param token: GitHub token with read_packages permission
    :param output_path: output file to write
    :returns: path to written Kubernetes config file
    """
    import json
    from base64 import b64encode
    output_path = output_path or join(getenv("NEON_CONFIG_PATH",
                                             "~/.config/neon"),
                                      f"k8s_secret_github.yml")
    output_path = expanduser(output_path)
    encoded_auth = b64encode(f"{username}:{token}".encode())
    auth_dict = {"auths": {"ghcr.io": {"auth": encoded_auth.decode()}}}
    auth_str = json.dumps(auth_dict)
    encoded_config = b64encode(auth_str.encode())
    secret_spec = {
        "kind": "Secret",
        "type": "kubernetes.io/dockerconfigjson",
        "apiVersion": "v1",
        "metadata": {
            "name": "github-auth"
        },
        "data": {
            ".dockerconfigjson": encoded_config.decode()
        }
    }
    with open(output_path, 'w+') as f:
        yaml.dump(secret_spec, f)
    return output_path


def _update_tcp_config(port_config: dict,
                       output_path: Optional[str] = None) -> str:
    """
    Generate (or update) a tcp-services config spec for ingress-nginx
    :param port_config: dict of "{port}" to "{namespace}/{service}:{port}"
    :param output_path: output file to write
    :returns: path to `tcp-services` k8s config map spec to be applied
    """
    # TODO: Deprecate Method
    LOG.warning("This function is deprecated")
    output_file = output_path or join(getenv("NEON_CONFIG_PATH",
                                             "~/.config/neon"), "ingress",
                                      f"k8s_config_tcp_services.yml")
    output_file = expanduser(output_file)

    if isfile(output_file):
        with open(output_file) as f:
            config = yaml.safe_load(f)
    else:
        config = {'apiVersion': 'v1',
                  'kind': 'ConfigMap',
                  'metadata': {
                      'name': 'tcp-services',
                      'namespace': 'ingress-nginx'
                  },
                  'data': {}
                  }
    config['data'] = {**config['data'], **port_config}
    with open(output_file, 'w+') as f:
        yaml.dump(config, f, default_flow_style=False)
    return output_file


def _patch_ingress_nginx_controller_service(name: str, port: int,
                                            target_port: Optional[int] = None,
                                            protocol: str = "TCP",
                                            output_path: Optional[str] = None):
    """
    Generate an updated nginx controller service spec
    :param name: name of TCP service to configure
    :param port: port to forward
    :param target_port: optional target port (else uses `port`)
    :param protocol: transport protocol (default TCP)
    :param output_path: path to output file to write
    """
    # TODO: Deprecate Method
    LOG.warning("This function is deprecated")
    output_file = output_path or join(getenv("NEON_CONFIG_PATH",
                                             "~/.config/neon"), "ingress",
                                      f"k8s_patch_nginx_service.yml")
    output_file = expanduser(output_file)
    if os.path.isfile(output_file):
        LOG.info(f"Reading ingress config from {output_file}")
        with open(output_file) as f:
            config = yaml.safe_load(f)
    else:
        config = {"spec": {"ports": []}}

    if any([i['name'] == name for i in config['spec']['ports']]):
        LOG.info(f"Skipping already configured port: {name}")
        return output_file
    config['spec']['ports'].append({'name': name,
                                    'port': port,
                                    'targetPort': target_port or port,
                                    'protocol': protocol})

    with open(output_file, 'w+') as f:
        yaml.dump(config, f, default_flow_style=False)
    return output_file


def _update_ingress_config(address: str, service: str, port: int,
                           cert_issuer: str = "letsencrypt-prod",
                           output_path: Optional[str] = None):
    """
    Update routing rules to route traffic to a service.
    Note this assumes you are using ingress-nginx and tls
    :param address: hostname rule applies to
    :param service: service name
    :param port: exposed container port service is running on
    :param cert_issuer: Name of k8s Issuer to provide TLS certificates
    :param output_path: path to output file to write
    """
    # TODO: Deprecate Method
    LOG.warning("This function is deprecated")
    output_file = output_path or join(getenv("NEON_CONFIG_PATH",
                                             "~/.config/neon"), "ingress",
                                      f"k8s_config_ingress.yml")
    output_file = expanduser(output_file)
    if os.path.isfile(output_file):
        LOG.info(f"Reading ingress config from {output_file}")
        with open(output_file) as f:
            config = yaml.safe_load(f)
        if config['metadata']['annotations']['cert-manager.io/issuer'] != cert_issuer:
            raise RuntimeError("Config cert_issuer conflict, skipping update")
    else:
        config = {
            "kind": "Ingress",
            "apiVersion": "networking.k8s.io/v1",
            "metadata": {
                "name": "ingress-diana",
                "annotations": {
                    "cert-manager.io/issuer": cert_issuer
                }
            },
            "spec": {
                "ingressClassName": "nginx",
                "tls": [
                    {"hosts": [],
                     "secretName": f"tls-{cert_issuer}"}
                ],
                "rules": []
            }
        }

    if address in config['spec']['tls'][0]['hosts']:
        LOG.warning(f"Skipping address already configured: {address}")
        return output_file

    config['spec']['tls'][0]['hosts'].append(address)
    config['spec']['rules'].append({"host": address,
                                    "http": {"paths": [
                                        {"path": '/',
                                         "pathType": "Prefix",
                                         "backend": {
                                             "service": {
                                                 "name": service,
                                                 "port": {
                                                     "number": port
                                                 }
                                             }
                                         }}
                                    ]}})
    with open(output_file, 'w+') as f:
        yaml.dump(config, f, default_flow_style=False)
    return output_file


def _create_cert_issuer(name: str, email: str,
                        output_path: Optional[str] = None) -> str:
    """
    Create a certificate Issuer k8s spec
    :param name: name for Issuer
    :param email: email to use for ACME registration
    :param output_path: spec file to write
    """
    # TODO: Deprecate Method
    LOG.warning("This function is deprecated")
    output_file = output_path or join(getenv("NEON_CONFIG_PATH",
                                             "~/.config/neon"), "ingress",
                                      f"k8s_config_cert_issuer.yml")
    output_file = expanduser(output_file)
    spec = {
        "apiVersion": "cert-manager.io/v1",
        "kind": "Issuer",
        "metadata": {
            "name": name
        },
        "spec": {
            "acme": {
                "server": "https://acme-v02.api.letsencrypt.org/directory",
                "email": email,
                "privateKeySecretRef": {
                    "name": name
                },
                "solvers": [
                    {
                        "http01": {
                            "ingress": {
                                "class": "nginx"
                            }
                        }
                    }
                ]
            }
        }
    }
    with open(output_file, 'w+') as f:
        yaml.safe_dump(spec, f, default_flow_style=False)
    return output_file



# NEON AI (TM) SOFTWARE, Software Development Kit & Application Development System
# All trademark and other rights reserved by their respective owners
# Copyright 2008-2021 Neongecko.com Inc.
# BSD-3
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from this
#    software without specific prior written permission.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS;  OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE,  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from enum import Enum, IntEnum
from os.path import join, dirname
import yaml


class Orchestrator(IntEnum):
    DOCKER = 0
    KUBERNETES = 1
    OPENSHIFT = 2


class ServiceClass(Enum):
    MQ = "mq-backend"
    HTTP = "http-backend"
    CORE = "neon-core"


def _get_services_by_class() -> dict:
    """
    Gets a dict of service names by ServiceClass
    """
    service_mappings = join(dirname(__file__), "templates", "service_mappings.yml")
    with open(service_mappings) as f:
        services = yaml.safe_load(f)
    # valid_service_classes = set([svc.get("service_class") for name, svc in services.items()])
    services_by_class = dict()
    for name, data in services.items():
        if data["service_class"] not in services_by_class:
            services_by_class[data["service_class"]] = list()
        services_by_class[data["service_class"]].append(name)
    return services_by_class


def valid_mq_services() -> set:
    """
    Get the set of valid MQ services that can be automatically configured
    """
    return set(_get_services_by_class()[ServiceClass.MQ.value])


def valid_http_services() -> set:
    """
    Get the set of valid HTTP services that can be automatically configured
    """
    return set(_get_services_by_class()[ServiceClass.HTTP.value])


def default_mq_services() -> set:
    """
    Return a specified set of default MQ services
    """
    return {"neon-rabbitmq",
            "neon-api-proxy",
            "neon-email-proxy",
            "neon-metrics-service"}



# NEON AI (TM) SOFTWARE, Software Development Kit & Application Development System
# All trademark and other rights reserved by their respective owners
# Copyright 2008-2021 Neongecko.com Inc.
# BSD-3
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from this
#    software without specific prior written permission.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS;  OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE,  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import yaml
import docker

from os import getenv, makedirs
from os.path import join, dirname, expanduser, isdir
from typing import Optional

from docker.errors import APIError
from docker.models.containers import Container
from ovos_utils.log import LOG


def run_clean_rabbit_mq_docker(bind_existing: bool = False) -> Container:
    """
    Start a clean RabbitMQ Docker instance to generate configuration files
    :param bind_existing: If true, allows for binding to a running container
        WARNING: setting `bind_existing` can result in overwriting valid MQ configuration
    """
    docker_client = docker.from_env()
    try:
        container = docker_client.containers.run("rabbitmq:3-management",
                                                 ports={"15672": "15672",
                                                        "5672": "5672"},
                                                 detach=True)
    except APIError as e:
        if e.status_code == 500:
            LOG.warning("Is an instance of RabbitMQ already running?")
            if bind_existing:
                for c in docker_client.containers.list():
                    if "rabbitmq:3-management" in c.image.tags:
                        LOG.info(f"Found a running RabbitMQ instance to configure")
                        return c
        else:
            LOG.error(e)
        raise e
    return container


def cleanup_docker_container(container_to_remove: Container):
    """
    Stop and remove the specified Docker Container
    :param container_to_remove: Docker.Container object to stop and remove
    """
    container_to_remove.stop()
    container_to_remove.remove()


def write_docker_compose(services_config: dict, compose_file: Optional[str] = None,
                         volume_type: str = "none",
                         volumes: Optional[dict] = None):
    """
    Generates and writes a docker-compose.yml file according to the specified services
    :param services_config: dict services, usually read from service_mappings.yml
    :param compose_file: path to docker-compose.yml file to write
    :param volume_type: volume type to use for config (default "none" uses local fs)
    :param volumes: Optional dict of volume names to directories (including hostnames for nfs volumes)
    """
    compose_file = compose_file if compose_file else \
        join(getenv("NEON_CONFIG_PATH", "~/.config/neon"), "docker-compose.yml")
    compose_file = expanduser(compose_file)

    with open(join(dirname(dirname(__file__)), "templates", "docker-compose.yml")) as f:
        compose_boilerplate = yaml.safe_load(f)
    compose_contents = {**compose_boilerplate, **{"services": services_config}}

    volumes = volumes or list()
    for vol in volumes:
        if volume_type == "nfs":
            volume_host, volume_path = volumes[vol].split(':')
            volume_opts = f"addr={volume_host},nolock,rw,soft,nfsvers=4"
            volume_path = f"\":{volume_path}\""
        else:
            volume_opts = "bind"
            volume_path = volumes[vol] or dirname(compose_file)
            if not isdir(volume_path):
                makedirs(volume_path)
        compose_contents["volumes"][vol] = {
            "driver_opts": {
                "type": f"{volume_type}-{vol}" if volume_type else vol,
                "o": volume_opts,
                "device": volume_path
            },
            "labels": {
                "kompose.volume.storage-class-name": f"nfs-{vol}"
            }
        }

    with open(compose_file, "w+") as f:
        yaml.dump(compose_contents, f)



# NEON AI (TM) SOFTWARE, Software Development Kit & Application Development System
# All trademark and other rights reserved by their respective owners
# Copyright 2008-2021 Neongecko.com Inc.
# BSD-3
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from this
#    software without specific prior written permission.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS;  OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE,  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import json
import shutil
import click
import yaml

from os import getenv, makedirs, walk, remove
from os.path import isdir, join, expanduser, isfile, dirname, abspath, exists
from pprint import pformat
from click_default_group import DefaultGroup
from ovos_utils.xdg_utils import xdg_config_home
from neon_diana_utils.constants import valid_mq_services, default_mq_services, \
    Orchestrator
from neon_diana_utils.version import __version__


@click.group("diana", cls=DefaultGroup,
             no_args_is_help=True, invoke_without_command=True,
             help="Diana: Device Independent API for Neon Applications.\n\n"
                  "See also: diana COMMAND --help")
@click.option("--version", "-v", is_flag=True, required=False,
              help='Print the current version')
def neon_diana_cli(version: bool = False):
    if version:
        click.echo(f"Diana version {__version__}")


# Backend
@neon_diana_cli.command(help="Configure a Diana Backend")
@click.option('--service', '-s', multiple=True,
              help="Optional service to configure")
@click.option('--default', '-d', is_flag=True, default=False,
              help="Configure the Default MQ Backend Services")
@click.option('--complete', '-c', '-a', is_flag=True, default=False,
              help="Configure All MQ Backend Services. (NOTE: This includes packages that require authentication)")
@click.option('--user', '-u', default="admin",
              help="Username to configure for administration")
@click.option('--password', '-p',
              help="Password associated with admin user to login to console")
@click.option('--volume-driver', default='none',
              help="Docker Volume Driver to use. 'none' and 'nfs' are currently supported")
@click.option('--volume-path', default=None,
              help="Optional fully-qualified path to config (i.e. /opt/diana, 192.168.1.10:/opt/diana")
@click.option('--skip-config', is_flag=True, default=False,
              help="Skip backend configuration and just generate orchestrator definitions")
@click.option('--namespace', '-n', default='default',
              help="Kubernetes namespace to configure services to run in")
@click.option('--mq-namespace', default=None,
              help="Kubernetes namespace to configure MQ services to run in")
@click.option('--http-namespace', default=None,
              help="Kubernetes namespace to configure HTTP services to run in")
@click.option('--http', is_flag=True, default=False,
              help="Configure HTTP backend services in addition to MQ")
@click.argument('config_path', default=getenv("NEON_CONFIG_DIR", "~/.config/neon/"))
def configure_backend(config_path, service, default, complete, user, password,
                      volume_driver, volume_path, skip_config, namespace, http,
                      mq_namespace, http_namespace):
    from neon_diana_utils.utils.backend import cli_configure_backend
    # Determine Services to Configure
    if complete:
        services_to_configure = valid_mq_services()
    elif default:
        services_to_configure = default_mq_services()
    else:
        services_to_configure = None
    if service:
        if services_to_configure:
            click.echo("Service set and individual services specified\n"
                       "Configuring specified services only.")
        services_to_configure = set(service)

    # Call setup
    import sys
    std_out = sys.stdout
    try:
        sys.stdout = open("/dev/null", 'a+')
        mq_namespace = mq_namespace or namespace
        http_namespace = http_namespace or namespace
        cli_configure_backend(config_path, services_to_configure, user,
                              password, http, volume_driver, volume_path,
                              skip_config, mq_namespace, http_namespace)
        sys.stdout = std_out
        if not skip_config:
            click.echo(f"Configured RabbitMQ Administrator: {user}")
            click.echo(f"Remember to place `ngi_auth_vars.yml` in {config_path}")
        click.echo(f"Configured API Services: {services_to_configure}")
        click.echo(f"Configuration was written to: {config_path}")
    except Exception as e:
        sys.stdout = std_out
        if not isinstance(e, ValueError):
            click.echo("An unexpected error occurred during configuration:")
        click.echo(e)
        click.echo("Call `diana configure-backend --help` for more info")


@neon_diana_cli.command(help="Start a Diana Backend")
@click.option('--attach', '-a', is_flag=True, default=False,
              help="Attach terminal to the started containers")
@click.option('--orchestrator', '-o', default="docker",
              help="Orchestrator (docker|kubernetes|openshift")
@click.argument('config_path', default=getenv("NEON_CONFIG_DIR", "~/.config/neon/"))
def start_backend(config_path, attach, orchestrator):
    from neon_diana_utils.utils.backend import cli_start_backend

    # Determine requested orchestrator
    if orchestrator == "docker":
        orchestrator = Orchestrator.DOCKER
    elif orchestrator == "kubernetes":
        orchestrator = Orchestrator.KUBERNETES
    elif orchestrator == "openshift":
        orchestrator = Orchestrator.OPENSHIFT
    else:
        click.echo(f"Invalid orchestrator specified: {orchestrator}")
        return

    # Start the backend
    click.echo(f"Starting backend")
    try:
        cli_start_backend(config_path, attach, orchestrator)
        if not attach:
            click.echo("Diana Backend Started")
    except Exception as e:
        if not isinstance(e, ValueError):
            click.echo("An unexpected error occurred during configuration:")
        click.echo(e)


@neon_diana_cli.command(help="Stop a Diana Backend")
@click.option('--orchestrator', '-o', default="docker",
              help="Orchestrator (docker|kubernetes|openshift")
@click.argument('config_path', default=getenv("NEON_CONFIG_DIR", "~/.config/neon/"))
def stop_backend(config_path, orchestrator):
    from neon_diana_utils.utils.backend import cli_stop_backend

    # Determine requested orchestrator
    if orchestrator == "docker":
        orchestrator = Orchestrator.DOCKER
    elif orchestrator == "kubernetes":
        orchestrator = Orchestrator.KUBERNETES
    elif orchestrator == "openshift":
        orchestrator = Orchestrator.OPENSHIFT
    else:
        click.echo(f"Invalid orchestrator specified: {orchestrator}")
        return

    try:
        cli_stop_backend(config_path, orchestrator)
        click.echo("Diana Backend Stopped")
    except Exception as e:
        if not isinstance(e, ValueError):
            click.echo("An unexpected error occurred during configuration:")
        click.echo(e)


# Kubernetes
@neon_diana_cli.command(help="Configure RabbitMQ and export user credentials")
@click.option("--username", "-u", help="RabbitMQ username")
@click.option("--password", "-p", help="RabbitMQ password")
@click.argument("output_path", default=None, required=False)
def configure_mq_backend(username, password, output_path):
    output_path = expanduser(output_path or join(xdg_config_home(), "diana"))
    if exists(output_path):
        click.echo(f"Path exists: {output_path}")
        return
    elif not isdir(dirname(output_path)):
        makedirs(dirname(output_path))

    # Get Helm charts in output directory for deployment
    shutil.copytree(join(dirname(__file__), "helm_charts"),
                    join(output_path))
    chart_path = join(output_path, "diana-backend")
    # Cleanup any leftover build files
    for root, _, files in walk(output_path):
        for file in files:
            if any((file.endswith(x) for x in (".lock", ".tgz"))):
                remove(join(root, file))
    try:
        from neon_diana_utils.utils.kubernetes_utils import \
            cli_make_github_secret
        from neon_diana_utils.utils.backend import generate_rmq_config, \
            generate_mq_auth_config
        rmq_config = generate_rmq_config()
        username = username or click.prompt("RabbitMQ Admin Username", type=str)
        password = password or click.prompt("RabbitMQ Admin Password", type=str,
                                            hide_input=True)

        rmq_config['users'].append({'name': username,
                                    'password': password,
                                    'tags': ['administrator']})
        with open(join(chart_path, "rabbitmq.json"), 'w+') as f:
            json.dump(rmq_config, f, indent=2)
        click.echo(f"Generated RabbitMQ config at {chart_path}/rabbitmq.json")
        mq_auth_config = generate_mq_auth_config(rmq_config)
        click.echo(f"Generated auth for services: {set(mq_auth_config.keys())}")
        if click.confirm("Configure GitHub token for private services?"):
            gh_username = click.prompt("GitHub username", type=str)
            gh_token = click.prompt("GitHub Token with `read:packages` "
                                    "permission", type=str)
            cli_make_github_secret(gh_username, gh_token, output_path)
            gh_secret_path = join(chart_path, "templates",
                                  "secret_gh_token.yaml")
            shutil.move(join(output_path, "k8s_secret_github.yml"),
                        gh_secret_path)
            click.echo(f"Generated GH secret at {gh_secret_path}")

        keys_config = _make_keys_config(False)

        mq_url = "neon-rabbitmq"
        mq_port = 5672

        # Assume ingress is configured as recommended
        # confirmed = False
        # mq_url = None
        # mq_port = None
        # while not confirmed:
        #     mq_url = click.prompt("MQ Service Name or URL", type=str,
        #                           default="neon-rabbitmq")
        #     mq_port = click.prompt("MQ Client Port", type=int, default=5672)
        #     click.echo(f"{mq_url}:{mq_port}")
        #     confirmed = click.confirm("Is this MQ Address Correct?")
        config = {**{"MQ": {"users": mq_auth_config,
                            "server": mq_url,
                            "port": mq_port}},
                  **keys_config}
        output_file = join(chart_path, "diana.yaml")
        click.echo(f"Writing configuration to {output_file}")
        with open(output_file, 'w+') as f:
            yaml.dump(config, f)
        click.echo(f"Helm charts generated in {output_path}")
    except Exception as e:
        click.echo(e)


@neon_diana_cli.command(help="Generate RabbitMQ definitions")
@click.argument("output_file", default=None, required=False)
def make_rmq_config(output_file):
    if isfile(output_file):
        click.echo(f"{output_file} already exists")
        return
    if not isdir(dirname(output_file)):
        makedirs(dirname(output_file))
    # TODO: Generate random passwords


@neon_diana_cli.command(help="Generate a configuration file with access keys")
@click.option("--skip-write", "-s", help="Skip writing config to file",
              is_flag=True)
@click.argument("output_file", default=None, required=False)
def make_keys_config(skip_write, output_file):
    _make_keys_config(not skip_write, output_file)


@neon_diana_cli.command(help="Generate Kubernetes secret for Github images")
@click.option("--username", "-u",
              help="Github username")
@click.option("--token", "-t",
              help="Github PAT with read_packages permission")
@click.argument('output_path', default=getenv("NEON_CONFIG_DIR", "~/.config/neon/"))
def make_github_secret(username, token, output_path):
    from neon_diana_utils.utils.kubernetes_utils import cli_make_github_secret
    try:
        output_path = cli_make_github_secret(username, token, output_path)
        click.echo(f"Generated outputs in {output_path}")
    except Exception as e:
        click.echo(e)


def _make_keys_config(write_config: bool, output_file: str = None):
    """
    Interactive configuration tool to prompt user for expected API keys and
    service accounts to be included in Configuration.
    :param write_config: If true, write config to `output_file`
    :param output_file: Configuration file to write keys to
    """
    if write_config:
        output_file = expanduser(abspath((output_file or join(xdg_config_home(),
                                                              "diana",
                                                              "diana.yaml"))))
        if isfile(output_file):
            click.echo(f"File already exists: {output_file}")
            return
        elif not isdir(dirname(output_file)):
            makedirs(dirname(output_file))

    api_services = dict()
    if click.confirm("Configure API Proxy Services?"):
        keys_confirmed = False
        while not keys_confirmed:
            wolfram_key = click.prompt("Wolfram|Alpha API Key", type=str)
            alphavantage_key = click.prompt("AlphaVantage API Key",
                                            type=str)
            owm_key = click.prompt("OpenWeatherMap API Key", type=str)
            api_services = {
                "wolfram_alpha": {"api_key": wolfram_key},
                "alpha_vantage": {"api_key": alphavantage_key},
                "open_weather_map": {"api_key": owm_key}
            }
            click.echo(pformat(api_services))
            keys_confirmed = click.confirm("Are these keys correct?")

    email_config = dict()
    if click.confirm("Configure Email Service?"):
        config_confirmed = False
        while not config_confirmed:
            email_addr = click.prompt("Email Address", type=str)
            email_password = click.prompt("Password", type=str)
            smtp_host = click.prompt("SMTP URL", type=str,
                                     default="smtp.gmail.com")
            smtp_port = click.prompt("SMTP Port", type=str,
                                     default="465")
            email_config = {"mail": email_addr,
                            "pass": email_password,
                            "host": smtp_host,
                            "port": smtp_port}
            click.echo(pformat(email_config))
            config_confirmed = \
                click.confirm("Is this configuration correct?")

    brands_config = dict()
    if click.confirm("Configure Brands/Coupons Service?"):
        config_confirmed = False
        while not config_confirmed:
            server_host = click.prompt("SQL Host Address", type=str,
                                       default="trackmybrands.com")
            sql_database = click.prompt("SQL Database", type=str,
                                        default="admintr1_drup1")
            sql_username = click.prompt("SQL Username", type=str)
            sql_password = click.prompt("SQL Password", type=str)
            brands_config = {"host": server_host,
                             "database": sql_database,
                             "user": sql_username,
                             "password": sql_password}
            click.echo(pformat(brands_config))
            config_confirmed = \
                click.confirm("Is this configuration correct?")

    chatgpt_config = dict()
    if click.confirm("Configure ChatGPT Service?"):
        config_confirmed = False
        while not config_confirmed:
            gpt_key = click.prompt("ChatGPT API Key", type=str)
            gpt_model = click.prompt("ChatGPT Model", type=str,
                                     default="gpt-3.5-turbo")
            gpt_role = click.prompt("ChatGPT Role", type=str,
                                    default="You are trying to give a short "
                                            "answer in less than 40 words.")
            gpt_context = click.prompt("ChatGPT Context depth", type=int,
                                       default=3)
            max_tokens = click.prompt("Maximum tokens in responses", type=int,
                                      default=100)
            chatgpt_config = {
                "key": gpt_key,
                "model": gpt_model,
                "role": gpt_role,
                "context_depth": gpt_context,
                "max_tokens": max_tokens
            }
            click.echo(pformat(chatgpt_config))
            config_confirmed = click.confirm("Is this configuration correct?")

    fastchat_config = dict()
    if click.confirm("Configure FastChat Service?"):
        config_confirmed = False
        while not config_confirmed:
            model = click.prompt("FastChat Model", type=str,
                                 default="fastchat")
            context = click.prompt("FastChat context depth", type=int,
                                   default=3)
            max_tokens = click.prompt("Max number of tokens in responses",
                                      type=int, default=128)
            num_processes = click.prompt(
                "Number of queries to handle in parallel",
                type=int, default=1)
            num_threads = click.prompt("Number of threads to use per query",
                                       type=int, default=2)
            fastchat_config = {
                "model": model,
                "context_depth": context,
                "max_tokens": max_tokens,
                "num_parallel_processes": num_processes,
                "num_threads_per_process": num_threads
            }
            click.echo(pformat(fastchat_config))
            config_confirmed = click.confirm("Is this configuration correct?")

    config = {"keys": {"api_services": api_services,
                       "emails": email_config,
                       "track_my_brands": brands_config},
              "ChatGPT": chatgpt_config,
              "FastChat": fastchat_config
              }
    if write_config:
        click.echo(f"Writing configuration to {output_file}")
        with open(output_file, 'w+') as f:
            yaml.dump(config, f)
    return config



# NEON AI (TM) SOFTWARE, Software Development Kit & Application Development System
# All trademark and other rights reserved by their respective owners
# Copyright 2008-2021 Neongecko.com Inc.
# BSD-3
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from this
#    software without specific prior written permission.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS;  OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE,  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import yaml
import json
import os
import itertools
import subprocess
import secrets

from os import makedirs, getenv
from os.path import expanduser, isdir, join, dirname, basename, exists, isfile
from typing import Optional, Set
from ovos_utils.log import LOG
from ovos_utils.xdg_utils import xdg_config_home
from ovos_utils.json_helper import merge_dict

from neon_diana_utils.constants import valid_http_services, Orchestrator
from neon_diana_utils.rabbitmq_api import RabbitMQAPI
from neon_diana_utils.utils.docker_utils import write_docker_compose
from neon_diana_utils.utils.kubernetes_utils import write_kubernetes_spec, \
    generate_secret


def cli_configure_backend(config_path: str, mq_services: Set[str],
                          admin_user: str, admin_pass: str, http: bool,
                          volume_driver: str, volume_path: str,
                          skip_config: bool, mq_namespace: str,
                          http_namespace: str):
    """
    Handle `configure-backend` CLI Command.
    """
    if not config_path:
        raise ValueError("config_path not specified")
    # Get actual config path
    config_path = expanduser(config_path)
    if isfile(config_path):
        raise FileExistsError(f"Specified output path is a file")
    if not isdir(config_path):
        makedirs(config_path)

    # Validate requested password
    if not skip_config and not admin_pass:
        raise ValueError("Null password")

    # Validate services
    if not mq_services and not http:
        raise ValueError("No services specified")

    # Parse volume paths
    volumes = {"config": join(volume_path, "config") if volume_path
               else config_path,
               "metrics": join(volume_path, "metrics") if volume_path
               else join(config_path, "metrics")}

    # Parse Namespaces
    namespaces = {"MQ_NAMESPACE": mq_namespace,
                  "HTTP_NAMESPACE": http_namespace}

    # Parse user and orchestrator configuration
    mq_services_config = _parse_services(mq_services, "mq-backend")
    users_to_configure, neon_mq_user_auth, \
        mq_docker_compose_configuration, mq_kubernetes_configuration = \
        _parse_configuration(mq_services_config)
    if http:
        _, _, http_docker_configuration, http_kubernetes_configuration = \
            _parse_configuration(_parse_services(valid_http_services(), "http-backend"))
        docker_configuration = {**mq_docker_compose_configuration,
                                **http_docker_configuration}
        kubernetes_configuration = [*mq_kubernetes_configuration,
                                    *http_kubernetes_configuration]
    else:
        docker_configuration = mq_docker_compose_configuration
        kubernetes_configuration = mq_kubernetes_configuration

    if not skip_config:
        configure_mq_backend(admin_user=admin_user, admin_pass=admin_pass,
                             services=mq_services_config,
                             users_config=users_to_configure,
                             neon_mq_user_auth=neon_mq_user_auth)

    generate_backend_config(docker_compose_config=docker_configuration,
                            kubernetes_config=kubernetes_configuration,
                            config_path=config_path,
                            volume_driver=volume_driver,
                            volumes=volumes,
                            namespaces=namespaces)


def cli_start_backend(config_path: str, attach: bool,
                      orchestrator: Orchestrator):
    """
    Start a backend specified at the passed config_path
    :param config_path: path to backend spec files
    :param attach: attach the calling thread to the command output
    :param orchestrator: orchestrator to use to start backend services
    """
    # TODO: Validate the requested file is for a backend
    if not config_path:
        raise ValueError("Null config_path")

    config_path = expanduser(config_path)
    if not isfile(join(config_path, "docker-compose.yml")):
        raise ValueError(f"docker-compose.yml not found in {config_path}")

    if orchestrator == Orchestrator.DOCKER:
        docker_compose_command = "docker-compose up"
        if not attach:
            docker_compose_command += " --detach"
            subprocess.Popen(["/bin/bash", "-c", f"cd {config_path} && {docker_compose_command}"]).communicate()
        else:
            subprocess.Popen(["/bin/bash", "-c", f"cd {config_path} && {docker_compose_command}"]).communicate()
    else:
        raise ValueError("The requested orchestrator is not currently supported")


def cli_stop_backend(config_path: str, orchestrator: Orchestrator):
    """
    Stop a backend specified at the passed config_path
    :param config_path: path to backend spec files
    :param orchestrator: orchestrator to use to start backend services
    """
    # TODO: Validate the requested file is for a backend
    config_path = expanduser(config_path)
    if not config_path:
        raise ValueError("Null config_path")
    if not isfile(join(config_path, "docker-compose.yml")):
        raise ValueError(f"docker-compose.yml not found in {config_path}")

    if orchestrator == Orchestrator.DOCKER:
        docker_compose_command = "docker-compose down"
        subprocess.Popen(["/bin/bash", "-c", f"cd {config_path} && {docker_compose_command}"]).communicate()
    else:
        raise ValueError("The requested orchestrator is not currently supported")


def write_neon_mq_config(credentials: dict, config_file: Optional[str] = None):
    """
    Takes the passed credentials and exports an MQ credential config file
    :param credentials: MQ User credentials
    :param config_file: `mq_config.json` file to write
    """
    # TODO: Deprecate Method
    LOG.warning("This function is deprecated")
    config_file = config_file if config_file else \
        join(getenv("NEON_CONFIG_PATH", "~/.config/neon"), "mq_config.json")
    config_file = expanduser(config_file)
    config_path = dirname(config_file)
    if not exists(config_path):
        os.makedirs(config_path)

    configuration = {"server": "neon-rabbitmq",
                     "users": credentials}
    LOG.info(f"Writing Neon MQ configuration to {config_file}")
    with open(config_file, 'w+') as new_config:
        json.dump(configuration, new_config, indent=2)

    # Generate k8s secret
    generate_secret("mq-config", {"mq_config.json": json.dumps(configuration)},
                    join(config_path, "config", "k8s_secret_mq-config.yml"))


def write_rabbit_config(api: RabbitMQAPI, config_file: Optional[str] = None):
    """
    Writes out RabbitMQ config files for persistence on next run
    :param api: Configured RabbitMQAPI object
    :param config_file: Path to `rabbit_mq_config.json` file to write
    """
    # TODO: Deprecate Method
    LOG.warning("This function is deprecated")
    config_file = config_file if config_file else \
        join(getenv("NEON_CONFIG_PATH", "~/.config/neon"),
             "rabbit_mq_config.json")
    config_file = expanduser(config_file)
    config_path = dirname(config_file)
    if not exists(config_path):
        os.makedirs(config_path)

    config = api.get_definitions()
    LOG.info(f"Exporting Rabbit MQ configuration to {config_file}")
    with open(config_file, "w+") as exported:
        json.dump(config, exported, indent=2)

    config_basename = basename(config_file)
    rmq_conf_contents = f"load_definitions = /config/{config_basename}"
    with open(join(config_path, "rabbitmq.conf"), 'w+') as rabbit:
        rabbit.write(rmq_conf_contents)


def _parse_services(requested_services: set,
                    service_class: str = "mq-backend") -> dict:
    """
    Parse requested services and return a dict mapping of valid service names
    to configurations read from service_mappings.yml
    :param requested_services: set of service names requested to be configured
    :param service_class: string class of services to parse (http, mq)
    :returns: mapping of service name to parameters required for configuration
    """
    # TODO: Deprecate Method
    LOG.warning("This function is deprecated")
    # Read configuration from templates
    template_file = join(dirname(dirname(__file__)), "templates",
                         "service_mappings.yml")
    with open(template_file) as f:
        template_data = yaml.safe_load(f)
    if not requested_services:
        return {}
    services_to_configure = {name: dict(template_data[name])
                             for name in requested_services if
                             name in template_data and
                             template_data[name]["service_class"] ==
                             service_class}

    # Warn for unknown requested services
    if set(services_to_configure.keys()) != set(requested_services):
        unhandled_services = [s for s in requested_services
                              if s not in services_to_configure.keys()]
        LOG.warning(f"Some requested services not handled: {unhandled_services}")
    return services_to_configure


def _parse_vhosts(services_to_configure: dict) -> set:
    """
    Parse MQ vhosts specified in the requested configuration
    :param services_to_configure: service mapping parsed from service_mappings.yml
    :returns: set of vhosts to be created
    """
    # TODO: Deprecate Method
    LOG.warning("This function is deprecated")
    default_vhosts = ["/neon_testing"]
    vhosts = [service.get("mq", service).get("mq_vhosts", [])
              for service in services_to_configure.values()]
    vhosts.append(default_vhosts)
    return set(itertools.chain.from_iterable(vhosts))


def _parse_configuration(services_to_configure: dict) -> tuple:
    # TODO: Deprecate Method
    LOG.warning("This function is deprecated")
    # Parse user and orchestrator configuration
    user_permissions = {
        "neon_api_utils": {"/neon_testing": {"read": "./*",
                                             "write": "./*",
                                             "configure": "./*"}}
    }
    neon_mq_auth = dict()
    docker_compose_configuration = dict()
    kubernetes_configuration = list()
    for name, service in services_to_configure.items():
        # Get service MQ Config
        if service.get("mq"):
            merge_dict(user_permissions,
                       service.get("mq", service).get("mq_user_permissions",
                                                      dict()))
            if service["mq"].get("mq_username"):
                # TODO: Update MQ services such that their service names match the container names DM
                neon_mq_auth[service.get("mq",
                                         service).get("mq_service_name", name)] = \
                    {"user": service.get("mq", service)["mq_username"]}
        docker_compose_configuration[name] = service["docker_compose"]
        kubernetes_configuration.extend(service.get("kubernetes") or list())
    return user_permissions, neon_mq_auth, docker_compose_configuration, kubernetes_configuration


def configure_mq_backend(admin_user: str, admin_pass: str,
                         config_path: str = None,
                         url: str = "http://0.0.0.0:15672",
                         services: dict = None,
                         users_config: dict = None,
                         neon_mq_user_auth: dict = None) -> dict:
    """
    Configure a new Diana RabbitMQ backend
    :param url: URL of admin portal (default=http://0.0.0.0:15672)
    :param admin_user: username to configure for RabbitMQ configuration
    :param admin_pass: password associated with admin_user
    :param config_path: local path to write configuration files
    :param services: dict of services to configure on this backend
    :param users_config: dict of user permissions to configure
    :param neon_mq_user_auth: dict of MQ service names to credentials
    :returns: dict MQ user configuration
    """
    # TODO: Deprecate Method
    LOG.warning("This function is deprecated")
    api = RabbitMQAPI(url)

    services = services or _parse_services({"neon-api-proxy",
                                            "neon-brands-service",
                                            "neon-email-proxy",
                                            "neon-script-parser",
                                            "neon-metrics-service"})
    if not (users_config and neon_mq_user_auth):
        users_config, neon_mq_user_auth, _, _ = \
            _parse_configuration(services)

    # Configure Administrator
    api.login(admin_user, admin_pass)

    # Configure vhosts
    vhosts_to_configure = _parse_vhosts(services)
    LOG.debug(f"vhosts={vhosts_to_configure}")
    for vhost in vhosts_to_configure:
        api.add_vhost(vhost)

    # Configure users
    LOG.debug(f"users={users_config}")
    credentials = api.create_default_users(list(users_config.keys()))
    api.add_user("neon_api_utils", "Klatchat2021")

    # Configure user permissions
    for user, vhost_config in users_config.items():
        for vhost, permissions in vhost_config.items():
            if not api.configure_vhost_user_permissions(vhost, user, **permissions):
                raise RuntimeError(f"Error setting Permission! {user} {vhost}")

    # Export and save rabbitMQ Config
    config_path = expanduser(config_path or join(xdg_config_home(), "diana"))
    rabbit_mq_config_file = join(config_path, "rabbit_mq_config.json")
    write_rabbit_config(api, rabbit_mq_config_file)

    # Write out MQ Connector config file
    for service in neon_mq_user_auth.values():
        service["password"] = credentials[service["user"]]
    neon_mq_config_file = join(config_path, "mq_config.json")
    write_neon_mq_config(neon_mq_user_auth, neon_mq_config_file)
    return neon_mq_user_auth


def generate_backend_config(docker_compose_config: dict,
                            kubernetes_config: list,
                            config_path: Optional[str] = None,
                            volume_driver: str = "none",
                            volumes: Optional[dict] = None,
                            namespaces: dict = None):
    """
    Generate orchestrator configuration for the specified services
    :param docker_compose_config: dict of Docker compose container specs
    :param kubernetes_config: list of Kubernetes Service/Deployment specs
    :param config_path: local path to write configuration files
           (default=NEON_CONFIG_PATH)
    :param volume_driver: Docker volume driver
           (https://docs.docker.com/storage/volumes/#use-a-volume-driver)
    :param volumes: Optional dict of volume names to directories
           (including hostnames for nfs volumes)
    :param namespaces: k8s namespaces to configure
    """
    # TODO: Deprecate Method
    LOG.warning("This function is deprecated")
    # Generate docker-compose file
    docker_compose_file = join(expanduser(config_path), "docker-compose.yml") if config_path else None
    write_docker_compose(docker_compose_config, docker_compose_file,
                         volume_driver, volumes)

    # Generate Kubernetes spec file
    write_kubernetes_spec(kubernetes_config, config_path, namespaces)


def generate_rmq_config() -> dict:
    """
    Generate a default configuration for RabbitMQ backend services
    """
    base_config_file = join(dirname(dirname(__file__)), "templates",
                            "rmq_backend_config.yml")
    with open(base_config_file) as f:
        base_config = yaml.safe_load(f)
    for user in base_config['users']:
        if user["password"]:
            # Skip users with defined passwords
            continue
        user['password'] = secrets.token_urlsafe(32)
    return base_config


def generate_mq_auth_config(rmq_config: dict) -> dict:
    """
    Generate an MQ auth config from RabbitMQ config
    :param rmq_config: RabbitMQ definitions, i.d. from `generate_rmq_config
    :returns: Configuration for Neon MQ-Connector
    """
    mq_user_mapping_file = join(dirname(dirname(__file__)), "templates",
                                "mq_user_mapping.yml")
    with open(mq_user_mapping_file) as f:
        mq_user_mapping = yaml.safe_load(f)

    mq_config = dict()
    LOG.debug(rmq_config.keys())
    for user in rmq_config['users']:
        username = user['name']
        for service in mq_user_mapping.get(username, []):
            mq_config[service] = {"user": username,
                                  "password": user['password']}
    return mq_config
